package cn.com.qmth.examcloud.core.reports.service.impl;

import java.util.Date;
import java.util.List;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.jdbc.core.JdbcTemplate;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import cn.com.qmth.examcloud.core.reports.base.util.online.ActiveDataUtil;
import cn.com.qmth.examcloud.core.reports.base.util.online.OnlineCount;
import cn.com.qmth.examcloud.core.reports.dao.ExamStudentCountRepo;
import cn.com.qmth.examcloud.core.reports.dao.entity.ExamStudentCountEntity;
import cn.com.qmth.examcloud.core.reports.service.ExamStudentCountService;

@Service
public class ExamStudentCountServiceImpl implements ExamStudentCountService {

    @Autowired
    private ExamStudentCountRepo examStudentCountRepo;

    @Autowired
    private JdbcTemplate jdbcTemplate;

    @Override
    public Long getOnlineCount(Long rootOrgId, Long orgId, Long examId) {
        StringBuffer sb = new StringBuffer();
        sb.append(" select sum(t.online_count) from ec_r_exam_student_count t where t.root_org_id=").append(rootOrgId);
        if (orgId != null) {
            sb.append(" and t.org_id=").append(orgId);
        }
        if (examId != null) {
            sb.append(" and t.exam_id=").append(examId);
        }
        Long count = jdbcTemplate.queryForObject(sb.toString(), Long.class);
        if (count == null) {
            count = 0l;
        }
        return count;
    }

    @Transactional
    @Override
    public synchronized void saveCount() {
        examStudentCountRepo.resetAllCount();
        List<OnlineCount> ret = ActiveDataUtil.getExamStudentCount(null);
        if (ret != null) {
            for (OnlineCount oc : ret) {
                addOrUpdateCount(oc);
            }
        }
    }

    private void addOrUpdateCount(OnlineCount oc) {
        ExamStudentCountEntity e = examStudentCountRepo.findByRootOrgIdAndExamId(oc.getRootOrgId(), oc.getExamId());
        if (e == null) {
            e = new ExamStudentCountEntity();
            e.setRootOrgId(oc.getRootOrgId());
            e.setExamId(oc.getExamId());
            Date now = new Date();
            e.setCreationTime(now);
            e.setUpdateTime(now);
        }
        e.setOnlineCount(oc.getOnlineCount());
        examStudentCountRepo.save(e);
    }

}
